#include "MatrixFactory_MxElasticity.h"
#include "ElementMatrixCalculator.h"
#include <iostream>
#include <utility>
#include <map>
#include <vector>
#include <iterator>
#include <dolfin.h>



using namespace dolfin;
using namespace std;

namespace pycc {

void printDofs(Dof_simplex_i& dof, bool glob2loc = false ){ 
    for (int i=0; i < dof.size(); i++) {
        std::cout <<"dof "<<i<<" "<<dof.glob_dof(i)<<std::endl; 
        if ( glob2loc) {
            vector<pair<int,int> > dof_vec; 
            dof_vec = dof.glob2loc(i); 
            vector<pair<int,int> >::const_iterator iter; 
            for (iter = dof_vec.begin(); iter != dof_vec.end() ; iter++) {  
                int e = (*iter).first;
                int j = (*iter).second; 
                std::cout <<"local e "<<e<<" local j "<<j<<" global i "<<dof.glob_dof(e,j)<<std::endl; 
            }


        }
    }
}


MatrixFactory:: MatrixFactory(Mesh* mesh_){
  mapfac = new MapMatrixFactory(mesh_); 
  // default order  
  order1 = 1; 
  order2 = 1; 
}

MatrixFactory:: ~MatrixFactory() {
  delete mapfac; 
}



FastMatSparse* MatrixFactory:: computeSigmaMassMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeSigmaMassMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  return new FastMatSparse(*(mapfac->matrix));  
}

FastMatSparse* MatrixFactory:: computePxSigmaMassMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computePxSigmaMassMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  jdof = mapfac->jdof; 
  return new FastMatSparse(*(mapfac->matrix));  
}

FastMatSparse* MatrixFactory:: computeUxDivSigmaMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeUxDivSigmaMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  jdof = mapfac->jdof; 
  return new FastMatSparse(*(mapfac->matrix));  
}






MapMatrixFactory:: ~MapMatrixFactory() {
  matrix->clear(); 
  rhs->clear(); 
  idof->clear(); 
  jdof->clear(); 

  delete matrix; 
  delete rhs; 
  delete idof; 
  delete jdof; 
}


MapMatrixFactory:: MapMatrixFactory(Mesh* mesh_) {
    mesh = mesh_;
    matrix = new map<pair<int,int>,double>();
    rhs = new map<int,double>(); 
    idof = new Dof_simplex_i(true,true); 
    jdof = new Dof_ptv(true,true); 
    order1 = 1; 
    order2 = 1; 
}

void dump_to_matlab(std::map<pair<int,int>,double> A, const char *varname, const char *s)
{
  ofstream outfile(s, ios::out);
  outfile << "%The matrix\n"; 
  std::map<std::pair<int,int>,double>::const_iterator Ait;
  //double eps = 1.0e-10; 

  for(Ait=A.begin(); Ait!=A.end(); Ait++){
    outfile << varname << "(" << (*Ait).first.first+1  << ","
            << (*Ait).first.second+1  << ")=" << (*Ait).second << ";\n"; 
  }
}


void MapMatrixFactory:: computeSigmaMassMatrix(bool append_to_matrix){ 

    if (! append_to_matrix) {
        matrix->clear();
        idof->clear(); 
        jdof->clear(); 
    }
    int element = -1; 

    if (mesh->numSpaceDim() == 2)  {
	    std::cout <<"Not implemented yet!"<<std::endl;
    }
    else if (mesh->numSpaceDim() == 3) {

        double p0[3]; 
        double p1[3]; 
        double p2[3]; 
        double p3[3]; 

        for (CellIterator cell(*mesh); !cell.end(); ++cell) {

            ++element;

            // Obtain vertices from dolfin mesh
            Vertex& v0 = (*cell).vertex(0);
            Vertex& v1 = (*cell).vertex(1);
            Vertex& v2 = (*cell).vertex(2);
            Vertex& v3 = (*cell).vertex(3);


            // Create double arrays with the data from the vertices
            p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
            p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
            p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
            p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 


            switch(order1) {
                case 1 :  matrix_factory_mass_sigma_3D_order1(*matrix,*idof,element,p0,p1,p2,p3); break; 
//                case 2 :  matrix_factory_mass_sigma_3D_order2(*matrix,*idof,element,p0,p1,p2,p3); break; 
//                case 3 :  matrix_factory_mass_sigma_3D_order3(*matrix,*idof,element,p0,p1,p2,p3); break; 
            }
        }
    }
//    printDofs(*idof); 
}

void MapMatrixFactory:: computePxSigmaMassMatrix(bool append_to_matrix){ 

    if (! append_to_matrix) {
        matrix->clear();
        idof->clear(); 
        jdof->clear(); 
    }
    int element = -1; 

    if (mesh->numSpaceDim() == 2)  {
	    std::cout <<"Not implemented yet!"<<std::endl;
    }
    else if (mesh->numSpaceDim() == 3) {

        double p0[3]; 
        double p1[3]; 
        double p2[3]; 
        double p3[3]; 

        for (CellIterator cell(*mesh); !cell.end(); ++cell) {

            ++element;

            // Obtain vertices from dolfin mesh
            Vertex& v0 = (*cell).vertex(0);
            Vertex& v1 = (*cell).vertex(1);
            Vertex& v2 = (*cell).vertex(2);
            Vertex& v3 = (*cell).vertex(3);


            // Create double arrays with the data from the vertices
            p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
            p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
            p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
            p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 


            switch(order1) {
                case 1 :  matrix_factory_mass_sigma_p_3D_order1(*matrix,*idof,*jdof,element,p0,p1,p2,p3); break; 
//                case 2 :  matrix_factory_mass_sigma_3D_order2(*matrix,*idof,element,p0,p1,p2,p3); break; 
//                case 3 :  matrix_factory_mass_sigma_3D_order3(*matrix,*idof,element,p0,p1,p2,p3); break; 
            }
        }
    }
//    printDofs(*idof); 
}

void MapMatrixFactory:: computeUxDivSigmaMatrix(bool append_to_matrix){ 

    if (! append_to_matrix) {
        matrix->clear();
        idof->clear(); 
        jdof->clear(); 
    }
    int element = -1; 

    if (mesh->numSpaceDim() == 2)  {
	    std::cout <<"Not implemented yet!"<<std::endl;
    }
    else if (mesh->numSpaceDim() == 3) {

        double p0[3]; 
        double p1[3]; 
        double p2[3]; 
        double p3[3]; 

        for (CellIterator cell(*mesh); !cell.end(); ++cell) {

            ++element;

            // Obtain vertices from dolfin mesh
            Vertex& v0 = (*cell).vertex(0);
            Vertex& v1 = (*cell).vertex(1);
            Vertex& v2 = (*cell).vertex(2);
            Vertex& v3 = (*cell).vertex(3);


            // Create double arrays with the data from the vertices
            p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
            p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
            p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
            p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 


            switch(order1) {
                case 1 :  matrix_factory_ux_div_sigma_3D_order1(*matrix,*idof,*jdof,element,p0,p1,p2,p3); break; 
//                case 2 :  matrix_factory_mass_sigma_3D_order2(*matrix,*idof,element,p0,p1,p2,p3); break; 
//                case 3 :  matrix_factory_mass_sigma_3D_order3(*matrix,*idof,element,p0,p1,p2,p3); break; 
            }
        }
    }
//    printDofs(*idof); 
}


#ifdef HAVE_ML_EPETRA

//-- Epetra --------------------------------------------------------------------------

EpetraMatrixFactory:: EpetraMatrixFactory(Mesh* mesh_, Epetra_Comm* Comm_){
    mapfac = new MapMatrixFactory(mesh_); 
    Comm = Comm_; 
    Map = new Epetra_Map(1, 0, *Comm); 
    order1 = 1; 
    order2 = 1; 
}

EpetraMatrixFactory:: ~EpetraMatrixFactory() {
    delete mapfac; 
    delete Map; 
}


Epetra_CrsMatrix* EpetraMatrixFactory:: computeSigmaMassMatrix(bool append_to_matrix) {
    mapfac->order1 = order1; 
    mapfac->computeSigmaMassMatrix(append_to_matrix); 
    idof = mapfac->idof; 
    delete Map; 
    Map = new Epetra_Map(idof->size(), 0, *Comm); 
    return map2epetra(*(mapfac->matrix),*Map);  
}

Epetra_CrsMatrix* EpetraMatrixFactory:: computePxSigmaMassMatrix(bool append_to_matrix) {
    mapfac->order1 = order1; 
    mapfac->computePxSigmaMassMatrix(append_to_matrix); 
    idof = mapfac->idof; 
    jdof = mapfac->jdof; 
    delete Map; 
    Map = new Epetra_Map(idof->size(), 0, *Comm); 
    return map2epetra(*(mapfac->matrix),*Map);  
}

Epetra_CrsMatrix* EpetraMatrixFactory:: computeUxDivSigmaMatrix(bool append_to_matrix) {
    mapfac->order1 = order1; 
    mapfac->computeUxDivSigmaMatrix(append_to_matrix); 
    idof = mapfac->idof; 
    jdof = mapfac->jdof; 
    delete Map; 
    Map = new Epetra_Map(idof->size(), 0, *Comm); 
    return map2epetra(*(mapfac->matrix),*Map);  
}





Epetra_CrsMatrix* map2epetra(map<pair<int,int>,double>& A_from, Epetra_Map& emap) { 

    std::map<std::pair<int,int>,double>::const_iterator Ait;

    Epetra_CrsMatrix* A_to = new Epetra_FECrsMatrix(Copy, emap,1); 


//    dump_to_matlab(A_from, "A", "A.m"); 

    double value; 
    vector<int> j_at_row; 
    vector<double> value_at_row; 
    Ait=A_from.begin();
    int curr_row= (*Ait).first.first;  
    int i, j; 
    int size; 

    while ( Ait != A_from.end() ) { 

        i = (*Ait).first.first;
        j = (*Ait).first.second;
        value = (*Ait).second; 

        if ( i == curr_row) { 

            j_at_row.push_back(j); 
            value_at_row.push_back(value); 

        } else { // new row! 


            // 1. Fill Epetra_CrsMatrix.
            if ( j_at_row.size() != value_at_row.size() ) { 
                //FIXME shold this be an exception ?? 
                std::cout << "This is most likely a bug in the function map2epetra" <<std::endl; 
            }

            size = j_at_row.size(); 

            // create the temporal pointer arrays Indices and Values 
            int* Indices =  new int[size]; 
            double* Values = new double[size];

            // fill Indices and Values 
            for (int d=0; d < size; d++) { 
                Indices[d] = j_at_row[d]; 
                Values[d]  = value_at_row[d];  
            }

            // insert previous row into Epetra_CrsMatrix
            A_to->InsertGlobalValues (curr_row, size, Values, Indices);


            // 2. Clear the structures i_at_row, j_at_row and value_at_row.
            j_at_row.clear(); 
            value_at_row.clear(); 

            // delete temporal pointer arrays 
            delete [] Indices; 
            delete [] Values;

            // 3. Insert the new values.
            j_at_row.push_back(j); 
            value_at_row.push_back(value); 

            curr_row = i; 
        }
        Ait++; 
    }

    // Finally, fill the last values 
    size = j_at_row.size(); 

    // create the temporal pointer arrays Indices and Values 
    int* Indices =  new int[size]; 
    double* Values = new double[size];

    // fill Indices and Values 
    for (int d=0; d < size; d++) { 
        Indices[d] = j_at_row[d]; 
        Values[d]  = value_at_row[d];  
    }

    // insert into Epetra_CrsMatrix
    A_to->InsertGlobalValues (curr_row, size, Values, Indices);


    // 2. Clear the structures i_at_row, j_at_row and value_at_row.
    j_at_row.clear(); 
    value_at_row.clear(); 

    // delete temporal pointer arrays 
    delete [] Indices; 
    delete [] Values;

    // 3. Insert the new values.
    j_at_row.push_back(j); 
    value_at_row.push_back(value); 

    A_to->FillComplete(); 

    return A_to; 

}

#endif //HAVE_ML_EPETRA 


}
