#
# This is an example file, showing how to generate ufc element code for some basic forms.
#

from swiginac import *

import os, sys

import SyFi

from ufccodegen.symbolic_utils import inner, mul, grad, div
from ufccodegen import MatrixForm, VectorForm, ScalarForm
from ufccodegen import compile_form, gen_form_code

import ufccodegen

# Set code generation options
#ufccodegen.options.include_from = "local" # default
#ufccodegen.options.include_from = "pycc"
ufccodegen.options.include_from = "SyFi"


# define some minimal and basic forms

def mass(u, v, G, Ginv):
    return inner(u, v)

def mass_with_c(u, v, c, G, Ginv):
    return inner( mul(c, u), v)

def stiffness(u, v, G, Ginv):
    Du = grad(u, Ginv)
    Dv = grad(v, Ginv)
    return inner(Du, Dv)

def stiffness_with_M(u, v, M, G, Ginv):
    Du = grad(u, Ginv)
    Dv = grad(v, Ginv)
    return inner( mul(M,Du), Dv )

def rhs(v, f, G, Ginv):
    return inner(f, v)

def L2norm(u, G, Ginv):
    return inner(u, u)

def H1norm(u, G, Ginv):
    Du = grad(u, Ginv)
    return inner(u, u) + inner(Du, Du)


def gen_basic_forms(nsd=2, order=1, mode='g'):
    # initialize SyFi, necessary for some differential operators
    SyFi.initSyFi(nsd)
    
    # define a single Lagrange element
    polygon = SyFi.newReferenceSimplex(nsd)
    fe      = SyFi.LagrangeFE(polygon, order)
    
    # wrap integrands in forms
    A  = MatrixForm(mass)
    B  = MatrixForm(mass_with_c)
    C  = MatrixForm(stiffness)
    D  = MatrixForm(stiffness_with_M)
    b  = VectorForm(rhs)
    L2 = ScalarForm(L2norm)
    H1 = ScalarForm(H1norm)
    
    # choose integration mode
    #itg_mode    = 'quadrature'
    quad_order  = 2*order
    itg_mode    = 'symbolic'
    
    # generate UFC code for these forms and elements
    if mode == 'c':
        print "Compiling extension modules"
        # compile the extension modules
        Ac  = compile_form(A,  fe, itg_mode, quad_order)
        Bc  = compile_form(B,  fe, itg_mode, quad_order)
        Cc  = compile_form(C,  fe, itg_mode, quad_order)
        Dc  = compile_form(D,  fe, itg_mode, quad_order)
        bc  = compile_form(b,  fe, itg_mode, quad_order)
        L2c = compile_form(L2, fe, itg_mode, quad_order)
        H1c = compile_form(H1, fe, itg_mode, quad_order)
        
        # now Ac, Bc, bc, L2c, H1c etc are ufc::element_tensor objects
        print Ac.rank(), bc.rank(), L2c.rank()
        # at this point, Dolfin or PyCC will be able to perform FEM assembly with Ac, Bc etc.
    elif mode == 'g':
        print "Writing generated code to files:"
        print gen_form_code(A,  fe, itg_mode, quad_order)
        print gen_form_code(B,  fe, itg_mode, quad_order)
        print gen_form_code(C,  fe, itg_mode, quad_order)
        print gen_form_code(D,  fe, itg_mode, quad_order)
        print gen_form_code(b,  fe, itg_mode, quad_order)
        print gen_form_code(L2, fe, itg_mode, quad_order)
        print gen_form_code(H1, fe, itg_mode, quad_order)
    else:
        print "Unknown mode", mode, ", c=compile, g=generate."

if __name__ == '__main__':
    if len(sys.argv) != 3:
        print "Usage: python", sys.argv[0], "{g|c} {syfi|local|pycc}"
    else:
        mode = sys.argv[1]
        ufccodegen.options.include_from = sys.argv[2]
        dims = [2]
        orders = [1]
        for d in dims:
            for o in orders:
                gen_basic_forms(d, o, mode)
