
class code_generator:
    """Base class for all code generators."""
    def __str__(self):
        return self.generate_header() + self.generate_code()
    def generate_header(self):
        raise "A code_generator must implement generate_header()!"
    def generate_code(self):
        raise "A code_generator must implement generate_code()!"



header_header = \
r"""%(comment)s

#ifndef __%(basename)s_H
#define __%(basename)s_H

%(includes)s

namespace %(namespace)s
{

"""


header_footer = \
"""

} // namespace

#endif
"""


code_header = \
r"""%(comment)s

%(includeheader)s

%(using)s

namespace %(namespace)s
{

"""


code_footer = \
"""

} // namespace
"""



UfcComment = \
"""// This code complies with UFC v.1.0rc1, and is generated with SyFi.
// 
// http://www.fenics.org/syfi/
"""


def as_list(x):
    if isinstance(x, list):
        return x
    if isinstance(x, tuple):
        return list(x)
    return [x]


import sre

def generate_code(basename, code_generators, system_includes, local_includes, comment=UfcComment, namespace="SyFi", using=[], do_include_header=True):
    """Generate code for a header file and a source file based
       on contributions from a list of code_generator objects.
       The source code will assume that basename.h is the name
       of the header file, and include this. This assumption can
       be turned off by the argument do_include_header."""
    # allow single valued arguments or lists
    code_generators   = as_list(code_generators)
    system_includes   = as_list(system_includes)
    local_includes    = as_list(local_includes)
    if using:
        print "In generate_code: Warning, argument 'using' is not implemented"
    # assemble code from all code generators
    header_body = "\n\n\n".join( [cg.generate_header() for cg in code_generators] )
    code_body   = "\n\n\n".join( [cg.generate_code()   for cg in code_generators] )
    # setup variables used in header and footer strings
    includeheader = do_include_header  and  ('#include "%s.h"' % (basename))  or  ""
    includes = "\n".join(["#include <%s>" % n for n in system_includes]) + "\n" \
             + "\n".join(['#include "%s"' % n for n in local_includes])
    vars = {'basename': basename, 'comment': comment, 'includes': includes, 'namespace': namespace, 'using': "", 'includeheader': includeheader }
    # combine code
    hcode  = (header_header % vars) + header_body + header_footer
    ccode  = (code_header % vars)   + code_body   + code_footer
    ccode = sre.sub('0+e', 'e', ccode)
    # done! return code
    return (hcode, ccode)



# ... Utilities for code generation

import swiginac

default_code_rule = r"double %(symbol)s = %(value)s;"
def gen_symbol_code(As, Av, prespace=" "*4, rule=default_code_rule):
    """As is a symbol or matrix with symbols, Av is a scalar or matrix of the same shape with the corresponding values.
       Generates code based on rule, default %s, for all elements of As and Av.""" % default_code_rule
    srule = prespace + rule
    def code_rule(s, v):
        return srule % { "symbol": s, "value": v }
    if type(As) == swiginac.matrix  or  type(As) == swiginac.lst:
        if (As.nops() != Av.nops()):
            raise "As and Av must have same size."
        code = "\n".join( [code_rule(As.op(i).printc(), Av.op(i).printc()) for i in xrange(As.nops())] )
    else:
        code = code_rule(As.printc(), Av.printc())
    return code + "\n"


def gen_symbol_declaration(As, prespace=" "*4, prefix="double ", postfix=";\n"):
    """As is a symbol or matrix or lst with symbols.
       Generates code for declaration of all symbols in As."""
    # wrap As in object with nops() and op(i)
    if not( type(As) == swiginac.matrix  or  type(As) == swiginac.lst ):
        m = swiginac.matrix(1,1)
        m.set(0,0,As)
        As = m
    # cat all symbols in comma separated string
    symbol_list = ", ".join([As.op(i).printc() for i in range(As.nops())])
    code = prespace + prefix + symbol_list + postfix
    return code


