import swiginac

from code_generator import code_generator
from affine_map import gen_geometry_code

import ufc

# FIXME:
# - test tabulate_facet_dofs
# - Reflect separate initialization and dof-searching from Dof_ptv in this code

# TODO:
# - Add support for using mesh entities for dof-mapping, possibly avoiding initialization
# - implement needs_mesh_entities


# swiginac.matrix(nsd, 1, dof_xi(fe,i))
def dof_xi(fe, i):
    """Return a swiginac column vector with the reference coordinates of dof i in fe,
        assuming ."""
    dofi = fe.dof(i)
    # check if the element is a scalar or vector element
    if isinstance(dofi[0], swiginac.numeric): 
        dofi0 = dofi
    elif isinstance(dofi[0], list):
        dofi0 = dofi[0]
    return dofi0 


class dof_map(code_generator):
    def __init__(self, fe):
        self.element         = fe
        self.name            = fe.str()
        self.classname       = "dof_map_" + self.name
        self.signature       = self.classname + " // generated by SyFi"
        # generate code components:
        self.vars = {
            'classname'           : self.classname,
            'constructor'         : '',
            'destructor'          : '',
            'members'             : self.gen_members(),
            'signature'           : '  return "%s";' % self.signature,
            'needs_mesh_entities' : self.gen_needs_mesh_entities(),
            'init_mesh'           : '  return true;',
            'init_cell'           : self.gen_init_cell(),
            'global_dimension'    : '  return dof.size();',
            'local_dimension'     : '  return %d;' % fe.nbf(),
            'tabulate_dofs'       : self.gen_tabulate_dofs(),
            'tabulate_facet_dofs' : self.gen_tabulate_facet_dofs(),
            'num_facet_dofs'      : self.gen_num_facet_dofs()
        }
    
    def generate_header(self):
        return ufc.dof_map_header % self.vars
    
    def generate_code(self):
        code = ufc.dof_map_implementation % self.vars
        return code + self.gen_members_source()
    
    def gen_members(self):
        return "    Dof_Ptv dof;\n";
    
    def gen_members_source(self):
        return ""
    
    def gen_init_cell(self):
        # TODO: Initialize Dof_Ptv, then update tabulate_dofs when this is working.
        code = """    %(classname)s * _this = const_cast<%(classname)s*>(this); // TODO: Fix this HACK!
    unsigned int *dofs = new unsigned int[local_dimension()];
    _this->tabulate_dofs(dofs, m, c);
    delete [] dofs;""" % self.__dict__
        return code
    
    def gen_needs_mesh_entities(self):
        # TODO: return false when a type of mesh entities are not needed, but this feature is currently not supported in UserMatrixFactory either.
        code = \
"""  switch(d)
  {
  case 0: return true; // vertices
  case 1: return true; // edges
  case 2: return true; // faces
  case 3: return true; // volumes
  }
  return false; // strange unsupported case or error"""
        return code
    
    def gen_tabulate_dofs(self):
        """Generates body for:
            void tabulate_dofs(unsigned int *dofs, const ufc::mesh& m, const ufc::cell& c) const
        """
        fe   = self.element
        nsd  = fe.get_polygon().no_space_dim()
        code = gen_geometry_code(nsd, detG=False)
        code += """

  //Dof_Ptv& local_dof = (Dof_Ptv&) dof;
  Dof_Ptv& local_dof = const_cast<Dof_Ptv&>(dof); // TODO: when initialization is done properly, and if the proper functions in Dof_ptv are made const, this is not necessary.

  // indices
  int top_dim = %d;
  int element = c.entity_indices[top_dim][0];

        """ % (nsd)
        # create some symbols
        G  = swiginac.symbolic_matrix(nsd, nsd, "G")
        x0 = swiginac.symbol("x0")
        y0 = swiginac.symbol("y0")
        z0 = swiginac.symbol("z0") 
        p0 = swiginac.matrix(nsd, 1, [x0,y0,z0][:nsd])
        
        for i in range(0, fe.nbf()):
            dofi = fe.dof(i)
            # check if the element is a scalar element
            if isinstance(dofi[0], swiginac.numeric): 
                dofm  = swiginac.matrix(nsd, 1, dofi)
                gdofm = G.mul(dofm).add(p0)
                
                code += "  double dof%d[%d];\n" % (i, nsd)
                for d in range(0,nsd):
                    code +=  "  dof%d[%d] = %s;" % (i, d, gdofm.op(d).printc())
                code += "\n  Ptv pdof%d(%d, dof%d);\n" % (i, nsd, i)
                code += "  dofs[%d] = local_dof.insert_dof(element, %d, pdof%d);\n\n" % (i, i, i)
            # check if the element is a vector element
            elif isinstance(dofi[0], list):
                dofi0 = dofi[0]
                dofi1 = dofi[1]
                dofm  = swiginac.matrix(nsd, 1, dofi0)
                gdofm = G.mul(dofm).add(p0)
                
                code  += "  double dof%d[%d];\n" % (i, nsd+1)
                for d in range(0, nsd):
                    code +=  "  dof%d[%d] = %s;" % (i, d, gdofm.op(d).printc())
                code += "\n  dof%d[%d] = %s;" % (i, nsd, dofi1.printc())
                code += "\n  Ptv pdof%d(%d, dof%d);\n" % (i, nsd+1, i)
                code += "  dofs[%d] = local_dof.insert_dof(element, %d, pdof%d);\n\n" % (i, i, i)
        return code
    
    def gen_num_facet_dofs(self):
        nsd = self.element.get_polygon().no_space_dim()
        nf  = nsd+1
        code = "    return %d;\n" % nf
        return code
    
    def gen_tabulate_facet_dofs(self):
        """Generate body for:
          virtual void tabulate_facet_dofs(unsigned int* dofs,
                                           const ufc::mesh& m,
                                           const ufc::cell& c,
                                           unsigned int facet) const;
        Should be general for elements with point evaluation dofs on simplices.
        """
        fe  = self.element
        nsd = fe.get_polygon().no_space_dim()
        nf  = nsd+1
        nbf = fe.nbf()
        # a list of dofs for each facet
        facet_dofs = [ [] for i in range(nf) ]
        # for all dofs
        for i in range(nbf):
            # get reference coordinates for dof i
            p = dof_xi(fe, i)
            # for all facets
            for j in range(nf):
                # if i is located on facet j
                if facet_check(j, p):
                    # append dof i to facet_dofs list for facet j 
                    facet_dofs[j].append(i)
        # generate code for each facet
        code = "  switch(facet)\n  {\n"
        for i in range(nf):
            dofs = facet_dofs[i]
            code += "  case %d:\n" % i
            code += "\n".join( ["    dofs[%d] = %d;" % (j, dofs[j]) for j in range(len(dofs))] )
            code += "\n    break;\n"
        code += "  default:\n"
        code += '    throw std::runtime_error("Invalid facet number.");\n'
        code += "  }\n"
        return code

# TODO: cell type checks
def facet_check(i, p):#, celltype):
    if len(p) == 1:
        #if celltype != 'interval': raise "Unknown celltype"
        x = p[0]
        if i == 0:
            return x == 0
        if i == 1:
            return x == 1
    if len(p) == 2:
        #if celltype != 'triangle': raise "Unknown celltype"
        x, y = p
        if i == 0:
            return x+y-1 == 0
        if i == 1:
            return x == 0
        if i == 2:
            return y == 0
    if len(p) == 3:
        #if celltype != 'tetrahedra': raise "Unknown celltype"
        x, y, z = p
        if i == 0:
            return x+y+z-1 == 0
        if i == 1:
            return x == 0
        if i == 2:
            return y == 0
        if i == 3:
            return z == 0
    # this seems to be general for 2d and 3d simplices:
    #if False:
    #    if i == 0: return sum(p)-1 == 0
    #    return p[i-1] == 0

