import swiginac
import SyFi
import operator
from code_generator import code_generator
from affine_map import gen_geometry_code

import ufc

# TODO:
# - vertex interpolation
# - mixed elements
# - quadrilaterals
# - dof evaluation for elements without point evaluation dofs


#def _as_matrix_type(v):
#    if isinstance(v, swiginac.matrix):
#        return v
#    else: 
#        m = swiginac.matrix(1,1)
#        m[0,0] = v
#        return m
    

class finite_element(code_generator):
    def __init__(self, fe):
        self.element           = fe
        self.name              = fe.str()
        self.classname         = "finite_element_" + self.name
        self.signature         = self.classname + " // generated by SyFi"
        self.space_dimension   = fe.nbf()
        # set cell shape  TODO: support for quadrilaterals must be enabled here, or get from SyFi::Polygon
        nsd = fe.get_polygon().no_space_dim()
        if False: # is_quadrilateral
            self.cell_shape = ("interval", "quadrilateral", "hexahedron")[nsd-1]
        else:
            self.cell_shape = ("interval", "triangle",      "tetrahedron")[nsd-1]
        # get value shape
        N0 = self.element.N(0)
        if isinstance(N0, swiginac.matrix):
            if (N0.rows() == 1) and (N0.cols() == 1):
                self.value_rank        = 0
                self.value_dimension   = (1,)
            elif (N0.rows() == 1) or (N0.cols() == 1):
                self.value_rank        = 1
                self.value_dimension   = (N0.nops(), )
            else:
                self.value_rank        = 2
                self.value_dimension   = (N0.rows(), N0.cols())
        else:
            self.value_rank        = 0
            self.value_dimension   = (1,)
        # generate the code components:
        self.vars = {
            "classname"                 : self.classname,
            "constructor"               : "",
            "destructor"                : "",
            "signature"                 : '  return "%s";'    % self.signature,
            "cell_shape"                : "  return ufc::%s;" % self.cell_shape,
            "space_dimension"           : "  return %d;"      % fe.nbf(),
            "value_rank"                : "  return %d;"      % self.value_rank,
            "value_dimension"           : self.gen_value_dimension(),
            "evaluate_basis"            : self.gen_evaluate_basis(),
            "evaluate_dof"              : self.gen_evaluate_dof(),
            "interpolate_vertex_values" : self.gen_interpolate_vertex_values(),
            "members"                   : self.gen_members(),
            "num_sub_elements"          : "  return 1;",
            "sub_element"               : "  return *this;",     # TODO: support mixed elements
        }
    
    def generate_header(self):
        return ufc.finite_element_header % self.vars
    
    def generate_code(self):
        code = ufc.finite_element_implementation % self.vars
        return code + self.gen_members_source()
    
    def gen_members(self):
        return "" # No data needed? Might add data members here to cache values in quadrature points etc.
    
    def gen_members_source(self):
        return "" # No data needed? Might add data members here to cache values in quadrature points etc.
    
    def gen_value_dimension(self):
        vr = self.value_rank
        vd = self.value_dimension
        if vr < 2:
            code = "  return %d;" % vd[0]
        else:
            code = "  switch(i)\n  {\n  %s\n  }" % \
                   "\n".join([ "  case %d: return %d;" % (i, vd[i]) for i in range(vr) ])
        return code
    
    def gen_evaluate_dof(self):
        """Generates:
           double evaluate_dof(unsigned int i, const ufc::function& f, const ufc::cell& c) const { ... }
           Currently generalized for all elements with point evaluation dofs."""
        # some useful variables
        fe  = self.element
        nsd = fe.get_polygon().no_space_dim()
        nbf = fe.nbf()
        value_size = reduce(operator.__mul__, self.value_dimension)
        # create some symbols
        G  = swiginac.symbolic_matrix(nsd, nsd, "G")
        x0 = swiginac.symbol("x0")
        y0 = swiginac.symbol("y0")
        z0 = swiginac.symbol("z0") 
        p0 = swiginac.matrix(nsd, 1, [x0,y0,z0][:nsd])
        # initial code
        code  = gen_geometry_code(nsd, detG=False)
        code += "  double v[%d];\n" % value_size
        code += "  double x[%d];\n" % nsd
        # skip the switch when only one basis function
        if nbf > 1: code += "  switch(i)\n  {\n"
        spaces = ""
        if nbf > 1: spaces = "  "
        # for all dofs
        for i in range(nbf):
            # check if the element is a scalar or vector element
            dofi = self.element.dof(i)
            if isinstance(dofi[0], swiginac.numeric):
                dof_xi_list = dofi
            elif isinstance(dofi[0], list):
                dof_xi_list = dofi[0]
            # point coordinates for this dof in reference coordinates
            dof_xi = swiginac.matrix(nsd, 1, dof_xi_list)
            # apply geometry mapping to get global coordinates
            dof_x  = (G.mul(dof_xi).add(p0)).evalm()
            # generate code to compute this point
            if nbf > 1: code += "  case %d:\n" % (i)
            for k in range(nsd):
                code += spaces + "  x[%d] = %s;\n" % (k, dof_x.op(k).printc())
            if nbf > 1: code += "    break;\n"
        if nbf > 1: code += "  }\n" # end switch
        # now we can evaluate the function
        code += "  f.evaluate(v, x, c);\n"
        code += "  return v[i %% %d];\n" % (value_size)
        return code
    
    def gen_interpolate_vertex_values(self):
        """Generates:
           void %(classname)s::interpolate_vertex_values(double* vertex_values, const double* nodal_values) const
           Currently not implemented."""
        # TODO: implement this
        fe  = self.element
        nbf = fe.nbf()
        polygon = fe.get_polygon()
        nsd = polygon.no_space_dim()
        nv  = polygon.no_vertices()
        nodal_values = swiginac.matrix(nbf, 1, [swiginac.symbol("nodal_values[%d]" % i) for i in range(nbf)])
        u = sum( [fe.N(i)*nodal_values.op(i) for i in range(nbf)] )
        #x = [polygon.coordinate(i) for i in range(nv)]
        #vertex_values = [u(x[i]) for i in range(nv)]
        #print "---------- vertex_values"
        #print vertex_values
        #print "---------- vertex_values"
        #code = symbols_to_code(vertex values) # TODO
        code = '    throw std::logic_error("FIXME: interpolate_vertex_values not implemented in generated code");'
        return code
    
    def gen_evaluate_basis(self):
        fe  = self.element
        nsd = fe.get_polygon().no_space_dim()
        nbf = fe.nbf()
        init_string = "\n".join(["  const double %s = coordinates[%d];" % (("x","y","z")[i],i) for i in range(nsd)])
        basis_string = ""
        if nbf>1: basis_string += "\n\n  switch(i)\n  {\n"
        for i in range(0, nbf):
            Ni = fe.N(i)
            if isinstance(Ni, swiginac.matrix):
                values = [Ni.op(d).evalf().printc() for d in range(0, Ni.nops())]
            else: 
                values = (Ni.evalf().printc(),)
            if nbf>1: basis_string += "  case %d:\n" % i
            basis_string += "\n".join([ "    values[%d] = %s;" % (d, values[d]) for d in range(len(values)) ])
            if nbf>1: basis_string += "\n    break;\n"
        if nbf>1: basis_string += "  }"
        return init_string + basis_string

