import swiginac

from code_generator import code_generator, gen_symbol_code, gen_symbol_declaration
from affine_map     import gen_geometry_code
from quadrature     import find_quadrature_rule, gen_quadrature_code
from finite_element import finite_element
from dof_map        import dof_map

import ufc

# TODO:
# - Boundary integration
# - Symbolic optimization
# - Find optimal quad_order from order of integrand polynomials



def permute(p):
    """Returns a permutation of all multiindices within the range of a rank 0, 1, or 2 tensor shape."""
    if len(p) == 0:
        return (0,)
    if len(p) == 1:
        return [(k,) for k in range(p[0])]
    if len(p) == 2:
        return [(k1,k2) for k1 in range(p[0]) for k2 in range(p[1])]
    raise "Rank 3 or higher not supported in permute()"


# TODO: test if optimization here has any effect on the speed,
#       otherwise keep it readable
def index_string(i, shape, optimize=False):
    """Constructs an index string for a row major (C type) indexing of a flattened tensor of rank 0, 1, or 2."""
    if len(shape) == 0:
        return "0"
    if len(shape) == 1:
        return "%d" % i[0]
    if not optimize:
        if len(shape) == 2:
            return "%d*%d + %d" % (shape[0], i[0], i[1])
    else:
        if len(shape) == 2:
            return "%d" % (shape[0] * i[0] + i[1])
    raise "Rank 3 or higher not supported in index_string()"


def _as_list_with_len(x, wantlen):
    """If is not a list type, it is repeated in a list wantlen times.
       Otherwise checks it x the correct length. Always returns a list
       with length wantlen or raises an exception."""
    if isinstance(x, tuple):
        x = list(x)
    if not isinstance(x, list):
        x = [x,]
    if len(x) == 1:
        x = wantlen*x
    if len(x) != (wantlen):
        raise "Got list with size " + str(len(x)) + ", need " + str(wantlen) + "."
    return x



class form(code_generator):
    """Code generator for an ufc form."""
    def __init__(self, name, rank, num_coefficients, fe_list,
                 interior_integrand=None, boundary_integrand=None, interior_boundary_integrand=None,
                 integration_mode='symbolic', quad_order=-1):
        """Initialize the form. Quad order < 0 means the order is chosen automatically (not completely implemented yet)."""
        # the callback functions that evaluate the integrand expressions
        self.interior_integrand          = interior_integrand
        self.boundary_integrand          = boundary_integrand
        self.interior_boundary_integrand = interior_boundary_integrand
        # number of arguments to the integrand functions
        self.rank               = rank
        self.num_coefficients   = num_coefficients
        # finite elements, if only one is provided it is expanded into a list
        self.fe_list            = _as_list_with_len(fe_list, num_coefficients+rank)
        # choice of integration mode
        self.integration_mode   = integration_mode
        self.quad_order         = quad_order
        # some strings
        self.name               = name
        postfix = name + "_" + self.fe_list[0].str()
        self.classname       = "form_"          + postfix
        self.cell_classname  = "cell_integral_" + postfix
        self.ext_classname   = "ext_integral_"  + postfix
        self.int_classname   = "int_integral_"  + postfix
        self.signature       = self.classname + " // generated by SyFi"
        # utility vars
        self.polygon         = self.fe_list[0].get_polygon()
        self.nsd             = self.polygon.no_space_dim()
        self.shape           = [ self.fe_list[i].nbf() for i in range(self.rank) ]
        self.i_perm          = permute(self.shape) # a permutation of all multiindices i for the element tensor Ai
        # generate code components
        self.cell_integral_vars = {
             'classname'       : self.cell_classname,
             'constructor'     : '',
             'destructor'      : '',
             'members'         : '',
             'tabulate_tensor' : self.gen_cell_tabulate_tensor()
            }
        self.exterior_facet_integral_vars = {
             'classname'       : self.ext_classname,
             'constructor'     : '',
             'destructor'      : '',
             'members'         : '',
             'tabulate_tensor' : self.gen_exterior_facet_tabulate_tensor()
            }
        self.interior_facet_integral_vars = {
             'classname'       : self.int_classname,
             'constructor'     : '',
             'destructor'      : '',
             'members'         : '',
             'tabulate_tensor' : self.gen_interior_facet_tabulate_tensor()
            }
        self.form_vars = {
             'classname'                       : self.classname,
             'constructor'                     : '',
             'destructor'                      : '',
             'members'                         : '',
             'rank'                            : '  return %d;'   % self.rank,
             'num_coefficients'                : '  return %d;'   % self.num_coefficients,
             'signature'                       : '  return "%s";' % self.signature,
             'create_dof_map'                  : self.gen_create_dof_map(),
             'create_finite_element'           : self.gen_create_finite_element(),
             'create_cell_integral'            : self.gen_create_cell_integral(),
             'create_exterior_facet_integral'  : self.gen_create_exterior_facet_integral(),
             'create_interior_facet_integral'  : self.gen_create_interior_facet_integral(),
            }
    
    def generate_header(self):
        """Generate class declaration for this form class."""
        # combine generated function components with code templates defined in the ufc module
        return ufc.form_header % self.form_vars
    
    def generate_code(self):
        """Generate class definition (the actual implementation) for this form class."""
        # combine generated function components with code templates defined in the ufc module
        code = ""
        if self.interior_integrand:
            vars = self.cell_integral_vars
            code += ufc.cell_integral_header % vars
            code += ufc.cell_integral_implementation % vars
            code += "\n"*3
        if self.boundary_integrand:
            vars = self.exterior_facet_integral_vars
            code += ufc.exterior_facet_integral_header % vars
            code += ufc.exterior_facet_integral_implementation % vars
            code += "\n"*3
        if self.interior_boundary_integrand:
            vars = self.interior_facet_integral_vars
            code += ufc.interior_facet_integral_header % vars
            code += ufc.interior_facet_integral_implementation % vars
            code += "\n"*3
        code += ufc.form_implementation % self.form_vars
        return code
    
    def create_dof_map(self, i):
        """Create a code generator for the dofmap of argument i."""
        return dof_map(self.fe_list[i])
    
    def create_finite_element(self, i):
        """Create a code generator for the finite element of argument i."""
        return finite_element(self.fe_list[i])
    
    def gen_create_dof_map(self):
        """Generate code for a factory function to create dof_map objects of correct types.
           The code for these classes must have been created outside for this code to compile."""
        code = "  switch(i)\n  {\n"
        for i in range(self.rank+self.num_coefficients):
            code += "  case %d:\n" % i
            code += "    return new %s();\n" % self.create_dof_map(i).classname
        code += "  }\n"
        code += '  throw std::runtime_error("Invalid index in create_dof_map()");'
        return code
    
    def gen_create_finite_element(self):
        """Generate code for a factory function to create finite_element objects of correct types.
           The code for these classes must have been created outside for this code to compile."""
        code = "  switch(i)\n  {\n"
        for i in range(self.rank+self.num_coefficients):
            code += "  case %d:\n" % i
            code += "    return new %s();\n" % self.create_finite_element(i).classname
        code += "  }\n"
        code += '  throw std::runtime_error("Invalid index in create_finite_element()");'
        return code
    
    def gen_create_cell_integral(self):
        """Generate factory function."""
        if self.interior_integrand:
            code = "  return new %s();" % self.cell_classname
        else:
            code = "  return 0;"
        return code
    
    def gen_create_exterior_facet_integral(self):
        """Generate factory function."""
        if self.boundary_integrand:
            code = "  return new %s();" % self.ext_classname
        else:
            code = "  return 0;"
        return code
    
    def gen_create_interior_facet_integral(self):
        """Generate factory function."""
        if self.interior_boundary_integrand:
            code = "  return new %s();" % self.int_classname
        else:
            code = "  return 0;"
        return code
    
    def tabulate_integrands(self, integrands_function):
        """Tabulate expressions for the integrands and eventual helper variables for each element tensor entry."""
        # (TODO: This function is probably not general enough for DG.)
        # define symbols for affine maps
        G    = swiginac.symbolic_matrix(self.nsd, self.nsd, "G")
        Ginv = swiginac.symbolic_matrix(self.nsd, self.nsd, "Ginv")
        # fill w with proper symbolic coefficients, using symbolic vectors with w[0][0] etc. as names
        w = []
        for j in range(self.num_coefficients):
            fe    = self.fe_list[self.rank+j]
            w_sym = [swiginac.symbol("w[%d][%d]" % (j,i)) for i in range(fe.nbf())]
            w    += [ sum([w_sym[i]*fe.N(i) for i in range(fe.nbf())]) ]
        # compute the integrands
        integrands = {}
        helpers    = {}
        for i in self.i_perm:
            v = [self.fe_list[j].N(i[j]) for j in range(self.rank)]
            integrands[i], he = integrands_function(v, w, G, Ginv)
            if he:
                helpers[i] = he
            #else:
            #    helpers[i] = []
        return integrands, helpers
    
    def gen_cell_tabulate_tensor(self):
        """Generate code to compute the contribution of a cell integral to the form."""
        # no interior contribution at all?
        if not self.interior_integrand:
            return "This line should never be compiled."
        
        # generate code for for affine maps
        code = gen_geometry_code(self.nsd, Ginv=True)
        
        # compute all integrand expressions
        integrands, helpers = self.tabulate_integrands(self.interior_integrand)
        
        # generate code for the integrals with the chosen integration method
        if self.integration_mode == 'quadrature'  or  self.integration_mode == 'q':
            code += self.gen_cell_tabulate_tensor_body_quadrature(integrands, helpers)
        elif self.integration_mode == 'symbolic'  or  self.integration_mode == 's':
            code += self.gen_cell_tabulate_tensor_body_symbolic(integrands, helpers)
        return code
    
    def gen_exterior_facet_tabulate_tensor(self):
        """Generate code to compute the contribution of a regular boundary integral to the form."""
        if not self.boundary_integrand:
            return "This line should never be compiled."
        code = '  throw std::runtime_error("TODO: exterior facet contribution not implemented.");'
        return code
    
    def gen_interior_facet_tabulate_tensor(self):
        """Generate code to compute the contribution of an interior facet integral to the form."""
        if not self.interior_boundary_integrand:
            return "This line should never be compiled."
        code = '  throw std::runtime_error("TODO: interior facet contribution not implemented.");'
        return code
    
    def gen_cell_tabulate_tensor_body_symbolic(self, interior_integrands, helpers):
        # perform symbolic integration of all integrands
        detG = swiginac.symbol("detG")
        A = dict( [(i, detG*self.polygon.integrate(interior_integrands[i])) for i in self.i_perm] )
        
        # TODO: In this function, we can do a lot of optimization, possibly adding more helper variables.
        #       F.ex. extracting subexpressions that are computed more than once.
        
        # TODO: Generate code for helpers here, but because of the integration only if they are independent of x,y,z.
        #       Otherwise, one could do automatic substitution of the variables like in the SyFi test.
        if helpers:
            print "ERROR: got helper variables from integrand, but using symbolic integration. This is not supported (yet)!"
        
        # generate code for each element tensor entry
        Aij_str  = "\n".join([ "  A[%s] = %s;" % (index_string(i, self.shape), A[i].evalf().printc()) for i in self.i_perm ])
        
        # ... and this code is our result
        return Aij_str
    
    def gen_cell_tabulate_tensor_body_quadrature(self, interior_integrands, helpers):
        pre_code   = "" # before the quadrature loop
        inner_code = "" # inside the quadrature loop
        
        # TODO: In this function, we can do a lot of optimization, possibly adding more helper variables.
        #       F.ex. extracting the parts of A that are not dependent on x,y,z,
        #       and extracting subexpressions that are computed more than once.
        
        # generate initialization code
        Aij_zero_str  = "  memset(A, 0, sizeof(double)*%d);\n" % (len(self.i_perm))
        pre_code     += Aij_zero_str
        
        # declare a complete and unique set of helper variables
        if helpers:
            declared_helper_symbols = {}
            decl_code = ""
            # NB! Assuming all integrands use the same set of helper symbols!
            for h in helpers[self.i_perm[0]]:
                decl_code += gen_symbol_declaration(h[0])
            pre_code += decl_code
        
        # generate code for each element tensor entry with helper variables
        A = interior_integrands
        defn_code_rule  = r"%(symbol)s = %(value)s;"
        for i in self.i_perm:
            helper_str = ""
            if helpers:
                helper_str = "\n"
                helperlist = helpers[i]
                for h in helperlist:
                    helper_str += gen_symbol_code(h[0], h[1], rule=defn_code_rule) + "\n"
            Aij_str = "    A[%s] += (%s) * quad_weight_detG;\n" % (index_string(i, self.shape), A[i].evalf().printc())
            inner_code += helper_str + Aij_str
        
        if self.quad_order < 0:
            # TODO: find optimal choice from order of integrand polynomial, by
            #       finding max degree of x,y,z in Aij for Aij in interior_integrands
            self.quad_order = 2
            #for ai in interior_integrands:
            #    ord = max_order(ai)
            #    if ord > self.quad_order:
            #        self.quad_order = ord
        quadrule  = find_quadrature_rule(order=self.quad_order, nsd=self.nsd)
        quad_code = gen_quadrature_code(quadrule, inner_code)
        
        # ... and this code is our result
        return pre_code + quad_code



def test_index_string():
    shape = tuple()
    i = tuple()
    print shape, "; ", i, "; ", index_string(i, shape)
    
    shape = (3,)
    i = (1,)
    print shape, "; ", i, "; ", index_string(i, shape)
    
    shape = (3,2)
    i = (1,2)
    print shape, "; ", i, "; ", index_string(i, shape)



if __name__ == '__main__':
    print "form.py, no test implemented"

