
class Form:
    """This class wraps an integrand function with a signature closer to the actual form
       in a way that is general and uniformly callable from the code generation tools."""
    def __init__(self, interior_integrand, boundary_integrand, name, rank, num_coefficients):
        self.i_integrand      = interior_integrand
        self.b_integrand      = boundary_integrand
        self.name             = name
        self.rank             = rank 
        self.num_coefficients = num_coefficients
    
    def integrand(self, v, w, G, Ginv, integr):
        """Meant to be called from the code generation tools.
           Note: len(v) == rank, len(w) == num_coefficients."""
        args = []
        if self.rank:
            args += v[:self.rank]
        if self.num_coefficients:
            args += w[:self.num_coefficients]
        args += [G, Ginv]
        return integr(*args)
    
    def has_interior_integrand(self):
        return self.i_integrand
    
    def has_boundary_integrand(self):
        return self.b_integrand
    
    def interior_integrand(self, v, w, G, Ginv):
        """Meant to be called from the code generation tools.
           Note: len(v) == rank, len(w) == num_coefficients."""
        integ = self.i_integrand
        itg = self.integrand(v, w, G, Ginv, integ)
        if hasattr(integ, 'helpers'):
            return (itg, integ.helpers)
        else:
            return (itg, {})
    
    def boundary_integrand(self, v, w, G, Ginv):
        """Meant to be called from the code generation tools.
           Note: len(v) == rank, len(w) == num_coefficients."""
        integ = self.b_integrand
        itg = self.integrand(v, w, G, Ginv, integ)
        if hasattr(integ, 'helpers'):
            return (itg, integ.helpers)
        else:
            return (itg, {})


class MatrixForm(Form): 
    def __init__(self, integrand, name=None, num_coefficients=-1):
        # TODO: this doesn't work for functors, only functions
        try:
            fc = integrand.func_code
            if num_coefficients < 0:
                num_coefficients = fc.co_argcount - 2 - 2
            if not name:
                name = fc.co_name
            print "Chose name = ", name, " num_coeffs = ", num_coefficients
        except:
            print "Failed to automagically determine num_coefficients and name."
        Form.__init__(self, integrand, None, name, 2, num_coefficients)

class VectorForm(Form): 
    def __init__(self, integrand, name=None, num_coefficients=-1):
        # TODO: this doesn't work for functors, only functions
        try:
            fc = integrand.func_code
            if num_coefficients < 0:
                num_coefficients = fc.co_argcount - 1 - 2
            if not name:
                name = fc.co_name
            print "Chose name = ", name, " num_coeffs = ", num_coefficients
        except:
            print "Failed to automagically determine num_coefficients and name."
        Form.__init__(self, integrand, None, name, 1, num_coefficients)

class ScalarForm(Form): 
    def __init__(self, integrand, name=None, num_coefficients=-1):
        # TODO: this doesn't work for functors, only functions
        try:
            fc = integrand.func_code
            if num_coefficients < 0:
                num_coefficients = fc.co_argcount - 0 - 2
            if not name:
                name = fc.co_name
            print "Chose name = ", name, " num_coeffs = ", num_coefficients
        except:
            print "Failed to automagically determine num_coefficients and name."
        Form.__init__(self, integrand, None, name, 0, num_coefficients)


# TODO: Implement action for matrix-free methods
#class MatrixFormAction(Form): 
#    def __init__(self, integrand, name="MatrixFormAction", num_coefficients=1):
#        Form.__init__(self, integrand, None, name, 1, num_coefficients)
#
#class VectorFormAction(Form): 
#    def __init__(self, integrand, name="VectorFormAction", num_coefficients=1):
#        Form.__init__(self, integrand, None, name, 0, num_coefficients)

