import swiginac
import SyFi as syfi


# ... Utilities for symbolic computations, mostly on matrices

import math

def symbol(name):
    """Returns a symbol from the SyFi symbol factory."""
    return syfi.get_symbol(name)

def symbolic_vector(n, name):
    """Returns a length n vector of symbols from the SyFi symbol factory.
       The symbols will be named 'name<i>' with i=0,...,n-1.
       If n > 10, <i> will be prepadded with zeros."""
    digits = 1 + int(math.log10(n-1))
    rule   = "%%s%%0%dd" % digits
    v      = swiginac.matrix(n, 1)
    for i in xrange(n):
        v.set(i,0, symbol(rule % (name,i)))
    return v

def symbolic_matrix(n, name):
    """Returns a n x n matrix of symbols from the SyFi symbol factory.
       The symbols will be named 'name<i><j>' with i,j=0,...,n-1."""
    A = swiginac.matrix(n, n)
    for i in xrange(n):
        for j in xrange(n):
            A.set(i,j, symbol("%s%d%d" % (name,i,j)))
    return A


def symbols(names):
    """Returns a list of symbols with names from the SyFi symbol factory.
       'names' must be a sequence of strings."""
    return [syfi.get_symbol(na) for na in names]

def symbolic_vectors(n, names):
    """Returns a list of length n vectors with symbols, see symbolic_vector(n, name).
       'names' must be a sequence of strings."""
    return [symbolic_vector(n, na) for na in names]

def symbolic_matrices(n, names):
    """Returns a list of n x n matrices with symbols, see symbolic_matrix(n, name).
       'names' must be a sequence of strings."""
    return [symbolic_matrix(n, na) for na in names]


def x(i):
    """Returns one of the spacial symbols x, y, z for i=0,...,2, and t for i=3."""
    return syfi.get_symbol( ("x", "y", "z", "t")[i] )

def t():
    """Returns the time symbol t."""
    return syfi.get_symbol("t")


def I(n):
    """Returns the n x n identity matrix."""
    return swiginac.unit_matrix(n)

def zeros(m, n):
    """Returns a m x n matrix with zeros."""
    A = swiginac.matrix(m, n)
    for i in xrange(m):
        for j in xrange(n):
            A.set(i,j, 0)
    return A

def ones(m, n):
    """Returns a m x n matrix with ones."""
    A = swiginac.matrix(m, n)
    for i in xrange(m):
        for j in xrange(n):
            A.set(i,j, 1)
    return A

def add(A, B):
    """Adds two swiginac expressions and calls evalm() on the result before returning."""
    return (A+B).evalm()

def sub(A, B):
    """Subtracts two swiginac expressions and calls evalm() on the result before returning."""
    return (A-B).evalm()

def mul(A, B):
    """Multiplies two swiginac expressions and calls evalm() on the result before returning."""
    return (A*B).evalm()

def is_indexed_type(A):
    """Checks if the argument is a matrix or lst."""
    return isinstance(A.evalm(), swiginac.matrix) or isinstance(A, swiginac.lst)

def inner(A, B):
    """Takes the inner product of A and B: multiplication for scalars, dot product for vectors and lsts, contraction for matrices."""
    A = A.evalm()
    B = B.evalm()
    if is_indexed_type(A) and is_indexed_type(B):
        if A.nops() != B.nops(): raise "Unmatching size of operands"
        return sum( [A.op(i)*B.op(i) for i in xrange(A.nops())] )
    return (A*B).evalm()

def contract(A,B):
    """See inner(A,B)."""
    return inner(A,B)

def det(A):
    """Returns the determinant of the argument."""
    return A.evalm().determinant()

def transpose(A):
    """Returns the transpose of the argument."""
    return swiginac.transpose(A.evalm())

def trace(A):
    """Returns the trace of the argument."""
    return swiginac.trace(A.evalm())

def diff(f, x):
    """Returns df/dx, where x is a symbol or symbolic matrix."""
    x = x.evalm()
    if type(x) == swiginac.matrix:
        A = zeros(x.rows(), x.cols())
        for i in xrange(A.rows()):
            for j in xrange(A.cols()):
                A[i,j] = diff(f, x[i,j])
        return A
    return swiginac.diff(f, x)

def grad(u, Ginv=None):
    """Returns the gradient of u w.r.t. the spacial variables x,y[,z], depending on the dimension specified by SyFi.initSyFi(nsd)."""
    if not (Ginv is None):
        return mul(Ginv.evalm().transpose(), syfi.grad(u.evalm()))
    return syfi.grad(u.evalm())

def div(u, Ginv=None):
    """Returns the divergence of u w.r.t. the spacial variables x,y[,z], depending on the dimension specified by SyFi.initSyFi(nsd)."""
    if not (Ginv is None):
        return inner(Ginv.evalm().transpose(), syfi.grad(u.evalm())) # FIXME: verify this!
    return syfi.div(u.evalm())

