#include "ExpressionSimplifierVisitor.h"

#include <sstream>
#include <iostream>

using namespace std;
using namespace GiNaC;
using namespace SyFi;



ExpressionSimplifierVisitor::ExpressionSimplifierVisitor(TempSymbolHandler & tsh_, ExpressionCollection & ec_):
    tsh(tsh_), ec(ec_)
{
}


void ExpressionSimplifierVisitor::simplify(symbol s, const ex & e)
{
    e.accept(*this);
    ec.insert_expression(s, result);
}


ex ExpressionSimplifierVisitor::simplify(const ex & e)
{
    e.accept(*this);
    return result; 
}

// helper function encapsulating the "result hack"
const ex & ExpressionSimplifierVisitor::doVisit(const ex & e)
{
    e.accept(*this);
    return result;
}


// ... visitors below:

// just treat any unhandled types as a single expression
void ExpressionSimplifierVisitor::visit(const basic & e)
{
    result = ec.register_expression(e);
}

// simplify arguments of functions
void ExpressionSimplifierVisitor::visit(const function & e)
{
    ex e2 = e;
    for(int i=0; i<e.nops(); i++)
    {
        ex arg = e.op(i);
        e2 = e2.subs( arg == doVisit(arg) );
    }
    result = ec.register_expression(e2);
}

// do nothing for the basic types
void ExpressionSimplifierVisitor::visit(const numeric & e)
{
    result = e;
}
void ExpressionSimplifierVisitor::visit(const constant & e)
{
    // TODO: evalf?
    result = e;
}
void ExpressionSimplifierVisitor::visit(const symbol & e)
{
    result = e;
}

// handles integer powers like a series of multiplications,
// and keeps other powers as power() objects
void ExpressionSimplifierVisitor::visit(const power & e)
{
    // first visit and simplify both subexpressions
    ex a, b; // a^b

    a = doVisit(e.op(0));
    b = doVisit(e.op(1));

    // if b is integer, treat as n multiplications of a
    ex c = b.evalf();
    if( is_a<numeric>(c) )
    {
        numeric nu = ex_to<numeric>(c);
        int n = (int) to_double(nu);
        if( is_zero(nu-n) )
        {
            // is an integer power, can simplify as a sequence of multiplications
            ex s = a;
            for(int i=0; i<n-1; i++)
            {
                s = ec.register_expression( s * a );
            }
            result = s;
            return;
        }
    }

    // not integer power, can't simplify more
    result = ec.register_expression(power(a,b));
}

// splits a sum into a series of single additions
void ExpressionSimplifierVisitor::visit(const add & e)
{
    ex s = doVisit(e.op(0));
    for(int i=1; i<e.nops(); i++)
    {
        s = ec.register_expression( s + doVisit(e.op(i)) );
    }
    result = s;
}

// splits a product into a series of single multiplications 
void ExpressionSimplifierVisitor::visit(const mul & e)
{
    ex s = doVisit(e.op(0));
    for(int i=1; i<e.nops(); i++)
    {
        s = ec.register_expression( s * doVisit(e.op(i)) );
    }
    result = s;
}

