#include "Robust.h" 
#include <fstream>
#include "tools.h"

using std::cout;
using std::endl;

namespace SyFi {

Robust:: Robust() : StandardFE() { 
  description = "Robust"; 
} 



Robust:: Robust(Polygon& p, int order) : StandardFE(p, order) { 
	compute_basis_functions(); 
} 


void Robust:: compute_basis_functions() {

    // remove previously computed basis functions and dofs
    Ns.clear(); 
    dofs.clear();

    order = 3; 

    if ( p == NULL ) {
        throw(std::logic_error("You need to set a polygon before the basisfunctions can be computed"));
    }



    if (p->str() == "ReferenceLine" || p->str() == "Line" ) { 

        cout <<"Can not define the Robust element on a line"<<endl;  

    } else if (p->str() == "ReferenceTriangle" || p->str() == "Triangle" ) { 


        description = "Robust_2D"; 

        Triangle& triangle = (Triangle&)(*p); 
        GiNaC::lst equations; 
        GiNaC::lst variables; 
        GiNaC::ex V_space = bernsteinv(2, order, triangle, "a"); 
        GiNaC::ex V = V_space.op(0); 
        GiNaC::ex V_vars = V_space.op(1); 

        GiNaC::ex divV = div(V); 
        ex_ex_map basis2coeff = pol2basisandcoeff(divV);   
        ex_ex_it iter; 


        // div constraints:  
        for (iter = basis2coeff.begin(); iter != basis2coeff.end(); iter++) {   
            GiNaC::ex basis = (*iter).first;  
            GiNaC::ex coeff= (*iter).second;  
            if ( coeff != 0 && ( basis.degree(x) > 0 || basis.degree(y)  > 0 )  ) { 
                equations.append( coeff == 0 ); 
            }
        }

        // constraints on edges: 
        for (int i=1; i<= 3; i++) {
            Line line = triangle.line(i); 
            GiNaC::symbol s("s"); 
            GiNaC::lst normal_vec = normal(triangle, i); 
            GiNaC::ex Vn = inner(V, normal_vec);   
            Vn = Vn.subs(line.repr(s).op(0)).subs(line.repr(s).op(1));
            basis2coeff = pol2basisandcoeff(Vn,s); 
            for (iter = basis2coeff.begin(); iter != basis2coeff.end(); iter++) {   
                GiNaC::ex basis = (*iter).first;  
                GiNaC::ex coeff= (*iter).second;  
                if ( coeff != 0 &&  basis.degree(s) > 1 )  
                { 
                    equations.append( coeff == 0 ); 
                }
            }
        }

        // dofs related to the normal on the edges  
        for (int i=1; i<= 3; i++) {
            Line line = triangle.line(i); 
            GiNaC::lst normal_vec = normal(triangle, i); 
            GiNaC::ex P1_space = bernstein(1, line, istr("a",i));
            GiNaC::ex P1 = P1_space.op(2); 
            GiNaC::ex Vn = inner(V, normal_vec);  

            GiNaC::ex basis; 
            for (int j=0; j< P1.nops(); j++) {
                basis = P1.op(j); 
                GiNaC::ex integrand = Vn*basis; 
                GiNaC::ex dofi =  line.integrate(integrand);  
//                dofs.insert(dofs.end(), dofi); 
                dofs.insert(dofs.end(), GiNaC::lst(line.vertex(0), line.vertex(1), j)); 
                GiNaC::ex eq = dofi == GiNaC::numeric(0); 
                equations.append(eq); 

            }
        }

        // dofs related to the tangent on the edges  
        for (int i=1; i<= 3; i++) {
            Line line = triangle.line(i); 
            GiNaC::lst tangent_vec = tangent(triangle, i); 
            GiNaC::ex P0_space = bernstein(0, line, istr("a",i));
            GiNaC::ex P0 = P0_space.op(2); 
            GiNaC::ex Vt = inner(V, tangent_vec);  

            GiNaC::ex basis; 
            for (int j=0; j< P0.nops(); j++) {
                basis = P0.op(j); 
                GiNaC::ex integrand = Vt*basis; 
                GiNaC::ex dofi =  line.integrate(integrand);  
//                dofs.insert(dofs.end(), dofi); 
                dofs.insert(dofs.end(), GiNaC::lst(line.vertex(0), line.vertex(1), 2)); 
                GiNaC::ex eq = dofi == GiNaC::numeric(0); 
                equations.append(eq); 

            }
        }


        variables = collapse(GiNaC::lst(V_vars.op(0),V_vars.op(1)));   

        GiNaC::matrix b; GiNaC::matrix A; 
        matrix_from_equations(equations, variables, A, b); 

        int ncols = A.cols();  
        GiNaC::matrix vars_sq(ncols, ncols);

        // matrix of symbols 
        for (unsigned r=0; r<ncols; ++r)
            for (unsigned c=0; c<ncols; ++c)
                vars_sq(r, c) = GiNaC::symbol();

        GiNaC::matrix id(ncols, ncols);

        // identity 
        const GiNaC::ex _ex1(1);
        for (unsigned i=0; i<ncols; ++i)
            id(i, i) = _ex1;


        // invert the matrix 
        GiNaC::matrix m_inv(ncols, ncols);
        m_inv = A.solve(vars_sq, id, GiNaC::solve_algo::gauss);


        for (int i=0; i<dofs.size(); i++) {
            b.let_op(11+i) = GiNaC::numeric(1); 
            GiNaC::ex xx = m_inv.mul(GiNaC::ex_to<GiNaC::matrix>(b)); 

            GiNaC::lst subs;  
            for (int ii=0; ii<xx.nops(); ii++) { 
                subs.append(variables.op(ii) == xx.op(ii)); 
            }
            GiNaC::ex Nj1 = V.op(0).subs(subs);   
            GiNaC::ex Nj2 = V.op(1).subs(subs);   
            Ns.insert(Ns.end(), GiNaC::matrix(2,1,GiNaC::lst(Nj1,Nj2))); 
            b.let_op(11+i) = GiNaC::numeric(0); 
        }
    }
}




} // namespace SyFi 


