#include <iostream>
#include <cassert>
#include <ginac/ginac.h>
#include <SyFi.h>

#include "ExpressionSimplifier.h"

using namespace std;
using namespace GiNaC;
using namespace SyFi;


// .................

// print code listing for a symbol/expression list
void genCodeSymbols(ostream & os, const list< pair<symbol,ex> > & slist )
{
	string pre("\tconst double ");
	string mid("\t= ");
	string post(";");
	for(list< pair<symbol,ex> >::const_iterator i  = slist.begin();
	                                            i != slist.end(); i++)
	{
		os << pre << csrc << i->first << mid << i->second << post << endl;
	}
}


inline void e2m(matrix & m, const ex & e)
{
	m = ex_to<matrix>(e.evalm());
}



// .................


// verify the correctness of the factorization
void assertExpression(const ex & orig, const symbol & res, const list< pair<symbol,ex> > & slist, const GiNaC::exmap & symToEx)
{
	// iterate over temporary symbols and calculate the result backwards
	ex r = res;
	for(int k=0; k<slist.size(); k++)
	{
		r = r.subs(symToEx);
	}
	// verify that the expressions match
	assert( is_zero(r-orig) );
}


// .................


// pick an expression to test
ex pickExpression(int i)
{
	ex x = get_symbol("x");
	ex y = get_symbol("y");
	ex z = get_symbol("z");

	ex e;
	switch(i)
	{
  case 0: e = 0; break;
	case 1: e = sin(x*x*y*y)*x*x*y*y*y;	break;
	case 2:	e = x;		break;
	case 3:	e = x+y+z;	break;
	case 4:	e = x*y*z;	break;
	case 5:	e = x*x*x*x*x*x*x*y*y*y*y*z*z*z;	break;
	case 6:
		e = power(x,3)*power(y,2) + power(x,2) + x*y*y;
		break;
	case 7:
		e = power(x,3)*power(y,2) + power(x,2) + x*y*y;
		e = power(e, e) + e;
		break;
	case 8:
		{
		e = 1;
		int cm1 = 6;	// complexity
		for(int i=0; i<cm1; i++)
		{
			for(int j=0; j<cm1; j++)
			{
				e = e + (i+1)*power(x,i)*(j+1)*power(y,j);
			}
		}
		break;
		}
	case 9:
		{
		ex c1 = get_symbol("c1");
		ex c2 = get_symbol("c2");
		ex D  = get_symbol("D");
        
		ex Jinv = symbolic_matrix(3,3,"Jinv");
		ex Jinv00 = Jinv.op(0);
		ex Jinv01 = Jinv.op(1);
		ex Jinv02 = Jinv.op(2);
		ex Jinv10 = Jinv.op(3);
		ex Jinv11 = Jinv.op(4);
		ex Jinv12 = Jinv.op(5);
		ex Jinv20 = Jinv.op(6);
		ex Jinv21 = Jinv.op(7);
		ex Jinv22 = Jinv.op(8);

		e = (3.31e-01*Jinv12*c2*Jinv10+3.31e-01*Jinv22*c2*Jinv21+3.31e-01*c1*Jinv02*Jinv00+3.31e-01*c2*(Jinv00*Jinv00)+1.6e-01*c1*(Jinv02*Jinv02)+3.31e-01*c1*Jinv02*Jinv01+3.31e-01*Jinv22*c2*Jinv20+3.31e-01*c2*(Jinv01*Jinv01)+1.6e-01*c2*(Jinv20*Jinv20)+6.63e-01*c2*Jinv01*Jinv00+1.6e-01*c2*(Jinv21*Jinv21)+3.31e-01*Jinv11*c2*Jinv10+6.63e-01*c2*Jinv02*Jinv00+3.31e-01*c2*Jinv21*Jinv20+1.6e-01*(Jinv22*Jinv22)*c2+6.63e-01*c2*Jinv02*Jinv01+1.6e-01*(Jinv11*Jinv11)*c2+3.31e-01*Jinv12*Jinv11*c2+3.31e-01*c1*Jinv01*Jinv00+3.31e-01*c2*(Jinv02*Jinv02)+1.6e-01*c1*(Jinv00*Jinv00)+1.6e-01*c2*(Jinv10*Jinv10)+1.6e-01*(Jinv12*Jinv12)*c2+1.6e-01*c1*(Jinv01*Jinv01))*D;
		}
	// add more tests with a new case here, and update the iteration limit in main
	default:
		e = 0;
	}
	return e;
}


int main(int argc, char **argv)
{
  bool verbose = false; // turn this off for the unit tests
  if(argc>1 && argv[1][0] == 'v') verbose = true;
	
	int from = 0;
	int to   = 9;
	for(int i=from; i<to; i++)
	{
		// pick an expression to test
		ex e = pickExpression(i);
		ex f = e*e;
    ExStats estat = count_ops(e);
    ExStats fstat = count_ops(f);
		symbol esy = get_symbol("e");
		symbol fsy = get_symbol("f");
    if(verbose)
    {
			   cout << endl << "Expression:         " << esy << " = " << e << ", flops = " << estat.flops << endl;
			   cout << endl << "Expression squared: " << fsy << " = " << f << ", flops = " << fstat.flops << endl;
    }
    else
    {
        cout << "Expression " << i << endl;
    }
		// find common factors
		ExpressionSimplifier es;
		es.add(esy, e);
		es.add(fsy, f);
		es.simplify();
    
		// generate code!
		//cout << "Variables:" << endl;
		//genCodeSymbols(cout, es.output.getSymExList());

		// verify correct results
		//cout << "Verifying correct results: ";

		ExpressionCollection & ec = const_cast<ExpressionCollection &>(es.get_output());
    int nsym = ec.get_symex_list().size();
    if(verbose) cout << "Number of symbols: " << nsym << endl;
    
		assertExpression(e, esy, ec.get_symex_list(), ec.get_sym_to_ex());
		cout << "1 ok, ";
		
    assertExpression(f, fsy, ec.get_symex_list(), ec.get_sym_to_ex());
		cout << "2 ok." << endl << endl;
	}

	return 0;
}

